open list nat decidable

namespace list

variables {α : Type} 
variables {X Y : list α} 
variables {x y : α}

def is_subseq : list α → list α → Prop
| []       _      := true   
| (x::X)   []     := false  
| (x::X)   (y::Y) := ((x=y) ∧ (is_subseq X Y)) 
                        ∨ ((is_subseq (x::X) Y))

def dis_subseq [decidable_eq α]: list α → list α → Prop
| []       _      := true   
| (x::X)   []     := false 
| (x::X)   (y::Y) := if x = y  then dis_subseq X Y
                               else dis_subseq (x::X) Y

local infix `⊴`  : 50 := is_subseq
local infix `⊴ᵈ` : 50 := dis_subseq

lemma nil_subseq (X : list α) : [] ⊴ X :=
begin
cases X with x X',
    {trivial},
    {trivial}
end

lemma nil_of_subseq {X : list α} (H : X ⊴ []) : X = [] :=
begin
cases X with x X',
    {refl},
    {apply false.elim H}
end

lemma subseq_nil (X : list α) (H : X = []) : X ⊴ [] :=
by {rw H, apply nil_subseq}

lemma subseq_nil_iff_nil (X : list α) : X ⊴ [] ↔ X = [] :=
begin
apply iff.intro,
    {apply nil_of_subseq},
    {intro h, rw h, apply nil_subseq}
end

lemma subseq_nil_of_subseq
    {X Y: list α} (H1 : X ⊴ Y) (H2 : Y ⊴ []) :
    X ⊴ [] :=
by {rw subseq_nil_iff_nil at H2, rw H2 at H1, apply H1} 

lemma subseq_of_subseq_nil 
    {X :list α} (Y : list α) (H : X ⊴ []) :
    X ⊴ Y :=
by {rw nil_of_subseq H, apply nil_subseq}

lemma length_of_subseq_nil {X : list α} (H : X ⊴ []) :
    length X = 0 :=
by {rw nil_of_subseq H, unfold length}

@[refl] lemma is_subseq.refl (X : list α) : X ⊴ X :=
begin
induction X with x X' IHX,
    {trivial},
    {left, apply and.intro (eq.refl x) (IHX)}
end

lemma subseq_self {X Y :list α} (H : X = Y) : X ⊴ Y :=
by rw H

lemma subseq_cons  (x : α) (X : list α) : X ⊴ x::X :=
begin
cases X with x X',
    {apply nil_subseq},
    {right, refl}
end

lemma length_le_of_subseq {X Y : list α} (H : X ⊴ Y) : 
    length X ≤ length Y := 
begin
revert X, induction Y with y Y' IHY,
    {intros X H, rw nil_of_subseq H},
    {intros X H, cases X with x X',
        {apply zero_le},
        {cases H with H_xyXY H_xXY,
            {apply succ_le_succ, apply IHY H_xyXY.right},
            {apply le_succ_of_le, apply IHY H_xXY}}}
end

lemma length_of_subseq
    {X Y : list α} (H1 : X ⊴ Y) (H2 : Y ⊴ X) :
     length X = length Y :=
begin
apply le_antisymm,
    {apply length_le_of_subseq H1},
    {apply length_le_of_subseq H2}
end

lemma eq_of_subseq
    {X Y : list α} (H1 : X ⊴ Y) (H2 : length X = length Y) :
    X = Y :=
begin
revert X, induction Y with y Y' IHY,
    {intros X H1 H2, apply nil_of_subseq H1},
    {intros X H1 H2, cases X with x X',
        {contradiction},
        {congr,
            {cases H1 with H_xyXY H_xXY,
                {apply and.left H_xyXY},
                {apply absurd H2, apply ne_of_lt,
                 apply lt_succ_of_le (length_le_of_subseq H_xXY)}},
            {cases H1 with H_xyXY H_xXY,
                {apply (IHY (and.right H_xyXY)(succ_inj H2))},
                {apply absurd H2, apply ne_of_lt,
                 apply lt_succ_of_le (length_le_of_subseq H_xXY)}}}}
end

lemma is_subseq.antisymm
    {X Y : list α} (H1 : X ⊴ Y) (H2 : Y ⊴ X) : X = Y :=
begin
apply eq_of_subseq H1, 
apply length_of_subseq H1 H2
end

@[trans]lemma is_subseq.trans 
    {X Y Z : list α} (H1 : X ⊴ Y) (H2 : Y ⊴ Z) : X ⊴ Z :=
begin
revert X Y, induction Z with z Z' IHZ,
    {intros X Y H1 H2, 
     apply subseq_nil_of_subseq H1 H2},
    {intros X Y H1 H2, cases Y with y Y',
        {apply subseq_of_subseq_nil (z::Z') H1},
        {cases X with x X',
            {apply nil_subseq},
            {cases H2 with H_yzYZ H_yYZ,    
                {cases H1 with H_xyXY H_xXY,
                    {left, apply and.intro,
                        {apply eq.trans H_xyXY.left H_yzYZ.left},
                        {apply IHZ H_xyXY.right H_yzYZ.right}},
                    {right, apply IHZ H_xXY H_yzYZ.right}},
                {right, apply IHZ H1 H_yYZ}}}}
end

instance is_subseq_is_partial_order : 
    is_partial_order (list α) is_subseq :=
{refl     := λ X, is_subseq.refl X,
 trans    := λ X Y Z, is_subseq.trans,
 antisymm := λ X Y, is_subseq.antisymm} 

lemma subseq_of_cons_subseq_cons
    (H : x :: X ⊴ x :: Y) : X ⊴ Y :=
begin
cases H with H_xyXY H_xXY,  
    {apply H_xyXY.right},
    {apply is_subseq.trans _ H_xXY, apply subseq_cons x}
end

lemma cons_subseq_cons
    (x : α) (H : X ⊴ Y) : x :: X ⊴ x :: Y :=
by {left, apply and.intro (eq.refl x) H}

lemma cons_subseq_cons_iff_subseq 
    (x : α) (X Y : list α) : x :: X ⊴ x :: Y ↔ X ⊴ Y :=
begin
apply iff.intro,
    {apply subseq_of_cons_subseq_cons},
    {apply cons_subseq_cons}
end

lemma subseq_cons_of_subseq 
    (y : α) {X Y : list α} (H_XY : X ⊴ Y) : X ⊴ y::Y :=
by apply trans H_XY (subseq_cons y Y)

lemma tail_subseq
    {x : α} {X Y : list α} (H_xXY : x::X ⊴ Y) : X ⊴ Y :=
by apply trans (subseq_cons x X) H_xXY

lemma tail_subseq_tail
    {x y : α} {X Y : list α} (H : (x::X) ⊴ (y::Y)) :
    X ⊴ Y :=
begin
cases H with H_xyXY H_xXY,
    {apply H_xyXY.right},
    {apply tail_subseq H_xXY}
end

lemma subseq_tail
    (H1 : (x::X) ⊴ (y::Y)) (H2 : ¬ x = y) :
    (x::X) ⊴ Y :=
begin
cases H1 with H_xyXY H_xXY,
    {cases H_xyXY with H_xy H_XY, contradiction},
    {apply H_xXY}
end

variable [decidable_eq α]

lemma dsubseq_of_subseq 
    {X Y : list α} (H : X ⊴ Y): X ⊴ᵈ Y :=
begin
revert X, induction Y with y Y' IHY,
    {intros X H, cases X with x X',
        {trivial},
        {apply false.elim H}},
    {intros X H, cases X with x X',
        {trivial},
        {cases decidable.em (x = y) with H_eq H_neq,
            {unfold dis_subseq, rw if_pos H_eq,
             cases H with H_xyXY H_xXY,
                {apply IHY H_xyXY.right},
                {apply IHY (tail_subseq H_xXY)}},
            {unfold dis_subseq, rw if_neg H_neq, 
             cases H with H_xyXY H_xXY,
                {cases H_xyXY with H_xy H_xXY, contradiction},
                {apply IHY H_xXY}}}}
end

lemma subseq_of_dsubseq
    {X Y : list α} (H : X ⊴ᵈ Y): X ⊴ Y :=
begin
revert X, induction Y with y Y' IHY,
    {intros X H, cases X with x X',
        {trivial},
        {apply false.elim H}},
    {intros X H, cases X with x X',
        {trivial},
        {cases decidable.em (x = y) with H_eq H_neq,
            {unfold dis_subseq at H, rw if_pos H_eq at H, 
             left, apply and.intro H_eq (IHY H)},
            {unfold dis_subseq at H, rw if_neg H_neq at H, 
             right, apply IHY H}}}
end

lemma subseq_iff_dsubseq (X Y : list α) :
    X ⊴ Y ↔ X ⊴ᵈ Y :=
begin
apply iff.intro,
    {apply dsubseq_of_subseq},
    {apply subseq_of_dsubseq}
end

lemma subseq_cons_iff (x y : α) (X Y : list α) :
    (x::X) ⊴ (y::Y) ↔ (x = y ∧ X ⊴ Y) ∨ (x ≠ y ∧ x::X ⊴ Y) :=
begin
apply iff.intro,
    {intro h, cases decidable.em (x = y) with heq hneq,
        {left, apply and.intro,
            {apply heq},
            {apply tail_subseq_tail h}},
        {right, apply and.intro,
            {apply hneq},
            {apply subseq_tail h hneq}}},
    {intro h, cases h,
        {left, apply h},
        {right, apply h.right}}
end

instance decidable_is_subseq (X : list α) : 
  decidable_pred (λ Y, is_subseq Y X) :=
begin
induction X with x X' IHX,
  {intro Y, cases Y with y Y',
    {simp, unfold is_subseq, apply_instance},
    {simp, unfold is_subseq, apply_instance}},
  {intro Y, cases Y with y Y',
    {simp, unfold is_subseq, apply_instance},
    {simp, unfold is_subseq, 
      have h1 : decidable (y = x ∧ is_subseq Y' X'), 
        from @and.decidable (y = x) (is_subseq Y' X') (by apply_instance) (IHX Y'),
      have h2 : decidable (is_subseq (y::Y') X'), from IHX (y::Y'),
      apply @or.decidable _ _ h1 h2}}
end

end list